#include <CL/sycl.hpp>
#include <CL/sycl/intel/fpga_extensions.hpp>

#include <iostream>

#include <tbb/tbb.h>
#include <tbb/task.h>

#include "utility.h"

#ifndef PARALLEL_SORT_STL_H
#define PARALLEL_SORT_STL_H

using namespace cl::sycl;

namespace internal
{
	std::size_t g_depth = 0L;

	template<class Container, class _Pred>
	void qsort3w(Container& array, std::size_t _First, std::size_t _Last, _Pred compare)
	{
		if (_First >= _Last) return;
		
		std::size_t _Size = array.size(); g_depth++;
		if (_Size > 0)
		{
			std::size_t _Left = _First, _Right = _Last;
			bool is_swapped_left = false, is_swapped_right = false;
			gen::ITEM _Pivot = array[_First];

			std::size_t _Fwd = _First + 1;
			while (_Fwd <= _Right)
			{
				if (compare(array[_Fwd], _Pivot))
				{
					is_swapped_left = true;
					std::swap(array[_Left], array[_Fwd]);
					_Left++; _Fwd++;
				}

				else if (compare(_Pivot, array[_Fwd])) {
					is_swapped_right = true;
					std::swap(array[_Right], array[_Fwd]);
					_Right--;
				}

				else _Fwd++;
			}

			tbb::task_group task_group;
			task_group.run([&]() {
				if (((_Left - _First) > 0) && (is_swapped_left))
					qsort3w(array, _First, _Left - 1, compare);
			});

			task_group.run([&]() {
				if (((_Last - _Right) > 0) && (is_swapped_right))
					qsort3w(array, _Right + 1, _Last, compare);
			});

			task_group.wait();
		}
	}

	template<class Container, class _Pred >
	void parallel_sort(Container& array, std::size_t _First, std::size_t _Last, _Pred compare)
	{
		g_depth = 0L;
		tbb::task_group task_group;
		task_group.run([&]() {
			internal::qsort3w(array, _First, _Last - 1, compare);
		});

		task_group.wait();
	}

	template<class Container, class _CompKey, class _CompVals>
	void parallel_stable_sort(Container& array, sycl::queue device_queue,
		_CompKey comp_key, _CompVals comp_vals)
	{
		std::vector<std::size_t> pv; 
		pv.resize(array.size()); pv[0] = 0;

		tbb::task_group task_group;
		task_group.run([&]() {
			internal::parallel_sort(array, 0L, array.size(), comp_key);
		});
		
		task_group.wait();
        
		cl::sycl::buffer<std::size_t> pv_buf{ pv.data(), array.size() };
		cl::sycl::buffer<gen::ITEM> array_buf{ array.data(), array.size() };

		device_queue.submit([&](sycl::handler& cgh) {
			auto pv_acc = pv_buf.get_access<cl::sycl::access::mode::write>(cgh);
			auto array_acc = array_buf.get_access<cl::sycl::access::mode::read>(cgh);
			cgh.parallel_for<class partition_kernel>(cl::sycl::range<1>{array.size() - 2},
				[=](cl::sycl::id<1> idx) {
					if ((comp_key(array_acc[idx[0]], array_acc[idx[0] + 1]) ||
						(comp_key(array_acc[idx[0] + 1], array_acc[idx[0]])))) {
						pv_acc[idx[0] + 1] = idx[0] + 1;
					}
				});
		}).wait();
        
        pv_buf.get_access<cl::sycl::access::mode::read>();

        auto _LastIt = std::remove_if(dpstd::execution::par_unseq,
            pv.begin() + 1, pv.end(), [&](const std::size_t index) { return index == 0L; });
        
        pv.resize(std::distance(pv.begin(), _LastIt));
		pv.push_back(std::distance(array.begin(), array.end()));
        
		tbb::parallel_for(tbb::blocked_range<std::size_t>(0, pv.size() - 1), \
			[&](const tbb::blocked_range<std::size_t>& r) {
				for (std::size_t index = r.begin(); index != r.end(); index++)
					internal::parallel_sort(array, pv[index], pv[index + 1], comp_vals);
			});
	}

	template<class BidirIt, class _CompKey, class _CompVals>
	void sequential_stable_sort(BidirIt _First, BidirIt _Last,
		_CompKey comp_key, _CompVals comp_vals)
	{
		std::sort(_First, _Last, comp_key);

		BidirIt	_First_p = _First, _Last_p = _First_p;
		for (BidirIt _FwdIt = _First + 1; _FwdIt != _Last; _FwdIt++)
		{
			if ((comp_key(*_FwdIt, *(_FwdIt - 1)) || (comp_key(*(_FwdIt - 1), *_FwdIt))))
			{
				_Last_p = _FwdIt;
				if (_First_p < _Last_p)
				{
					std::sort(_First_p, _Last_p, comp_vals);
					_First_p = _Last_p;
				}
			}
		}

		std::sort(_First_p, _Last, comp_vals);
	}
}

#endif // PARALLEL_SORT_STL_H
